# Copyright (c) 2016 OpenStack Foundation
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from copy import deepcopy
from os import path as os_path
from string import Template
from unittest import mock

import fixtures
from oslo_serialization import jsonutils
from requests.exceptions import HTTPError

from neutron.db import provisioning_blocks
from neutron.plugins.ml2 import driver_context as ctx
from neutron.plugins.ml2 import plugin as ml2_plugin
from neutron.tests.unit.db import test_db_base_plugin_v2 as test_plugin
from neutron.tests.unit import testlib_api
from neutron_lib.api.definitions import portbindings
from neutron_lib.callbacks import events
from neutron_lib.callbacks import registry
from neutron_lib.callbacks import resources
from neutron_lib import constants as n_const
from neutron_lib import fixture
from neutron_lib.plugins import constants as plugin_constants
from neutron_lib.plugins import directory
from neutron_lib.plugins.ml2 import api
from oslo_config import fixture as config_fixture

from networking_odl.common import odl_features
from networking_odl.common import websocket_client
from networking_odl.journal import periodic_task
from networking_odl.ml2 import pseudo_agentdb_binding
from networking_odl.tests import base


AGENTDB_BINARY = 'neutron-odlagent-portbinding'
L2_TYPE = "ODL L2"
# test data hostconfig and hostconfig-dbget
SAMPLE_ODL_HCONFIGS = {"hostconfigs": {"hostconfig": [
    {"host-id": "devstack",
     "host-type": "ODL L2",
     "config": """{"supported_vnic_types": [
                {"vnic_type": "normal", "vif_type": "ovs",
                 "vif_details": {}}],
                "allowed_network_types": [
                "local", "vlan", "vxlan", "gre"],
                "bridge_mappings": {"physnet1": "br-ex"}}"""}
]}}


class OpenDaylightAgentDBFixture(fixtures.Fixture):
    def _setUp(self):
        super(OpenDaylightAgentDBFixture, self)._setUp()
        fake_agents_db = mock.MagicMock()
        fake_agents_db.create_or_update_agent = mock.MagicMock()
        self.useFixture(fixture.PluginDirectoryFixture())
        directory.add_plugin(plugin_constants.CORE, fake_agents_db)


class TestPseudoAgentDBBindingTaskBase(base.DietTestCase):
    """Test class for AgentDBPortBindingTaskBase."""

    def setUp(self):
        """Setup test."""
        self.useFixture(base.OpenDaylightRestClientFixture())
        self.useFixture(base.OpenDaylightPseudoAgentPrePopulateFixture())
        self.useFixture(OpenDaylightAgentDBFixture())
        super(TestPseudoAgentDBBindingTaskBase, self).setUp()

        self.worker = pseudo_agentdb_binding.PseudoAgentDBBindingWorker()
        self.task = pseudo_agentdb_binding.PseudoAgentDBBindingTaskBase(
            self.worker)

    def _get_raised_response(self, json_data, status_code):

        class MockHTTPError(HTTPError):
            def __init__(self, json_data, status_code):
                self.json_data = json_data
                self.status_code = status_code
                self.response = self

        class MockResponse(object):
            def __init__(self, json_data, status_code):
                self.raise_obj = MockHTTPError(json_data, status_code)

            def raise_for_status(self):
                raise self.raise_obj

        return MockResponse(json_data, status_code)

    def test_hostconfig_response_404(self):
        with mock.patch.object(self.task.odl_rest_client,
                               'get', return_value=self.
                               _get_raised_response({}, 404)):
            self.assertEqual(self.task._rest_get_hostconfigs(), [])


class TestPseudoAgentDBBindingPrePopulate(base.DietTestCase):
    KNOWN_HOST = 'known_host'
    AGENT_TYPE = pseudo_agentdb_binding.PseudoAgentDBBindingWorker.L2_TYPE

    def setUp(self):
        self.useFixture(base.OpenDaylightRestClientFixture())
        self.useFixture(OpenDaylightAgentDBFixture())
        super(TestPseudoAgentDBBindingPrePopulate, self).setUp()
        self.useFixture(fixture.CallbackRegistryFixture())

        self.ml2_plugin = mock.Mock()
        self.ml2_plugin.get_agents = mock.Mock(return_value=[])
        self.worker = mock.Mock()
        self.worker.known_agent = mock.Mock(return_value=False)
        self.worker.add_known_agent = mock.Mock()
        self.worker.update_agetns_db_row = mock.Mock()
        self.prepopulate = (pseudo_agentdb_binding.
                            PseudoAgentDBBindingPrePopulate(self.worker))

    def _call_before_port_binding(self, host):
        registry.publish(resources.PORT, events.BEFORE_CREATE, self.ml2_plugin,
                         payload=events.DBEventPayload(
                             mock.Mock(),
                             states=({portbindings.HOST_ID: host},)))

    def test_unspecified(self):
        self._call_before_port_binding(n_const.ATTR_NOT_SPECIFIED)
        self.worker.known_agent.assert_not_called()

    def test_empty_host(self):
        self._call_before_port_binding('')
        self.worker.known_agent.assert_not_called()

    def test_known_agent(self):
        self.worker.known_agent = mock.Mock(return_value=True)
        self._call_before_port_binding(self.KNOWN_HOST)
        self.worker.known_agent.assert_called()
        self.ml2_plugin.get_agents.assert_not_called()

    def test_agentdb_alive(self):
        self.ml2_plugin.get_agents = mock.Mock(return_value=[
            {'host': self.KNOWN_HOST,
             'agent_type': self.AGENT_TYPE,
             'alive': True}])
        self._call_before_port_binding(self.KNOWN_HOST)
        self.worker.known_agent.assert_called()
        self.ml2_plugin.get_agents.assert_called()
        self.worker.add_known_agents.assert_called_with([
            {'host': self.KNOWN_HOST,
             'agent_type': self.AGENT_TYPE,
             'alive': True}])
        self.worker.update_agents_db_row.assert_not_called()

    def test_agentdb_dead(self):
        self.ml2_plugin.get_agents = mock.Mock(return_value=[
            {'host': self.KNOWN_HOST,
             'agent_type': self.AGENT_TYPE,
             'alive': False}])
        self._call_before_port_binding(self.KNOWN_HOST)
        self.worker.known_agent.assert_called()
        self.ml2_plugin.get_agents.assert_called()
        self.worker.add_known_agents.assert_not_called()

    def test_unkown_hostconfig(self):
        with mock.patch.object(self.prepopulate,
                               'odl_rest_client') as mock_rest_client:
            mock_response = mock.Mock()
            mock_response.json = mock.Mock(
                return_value=SAMPLE_ODL_HCONFIGS['hostconfigs'])
            mock_rest_client.get = mock.Mock(return_value=mock_response)
            self._call_before_port_binding(self.KNOWN_HOST)
            self.worker.known_agent.assert_called()
            self.ml2_plugin.get_agents.assert_called()
            self.worker.add_known_agent.assert_not_called()
            self.worker.update_agents_db_row.assert_called_once()

    def test_http_error(self):
        with mock.patch.object(self.prepopulate,
                               'odl_rest_client') as mock_rest_client:
            mock_rest_client.get = mock.Mock(side_effect=Exception('error'))
            self._call_before_port_binding(self.KNOWN_HOST)
            self.worker.known_agent.assert_called()
            self.ml2_plugin.get_agents.assert_called()
            self.worker.add_known_agent.assert_not_called()
            self.worker.update_agents_db_row.assert_not_called()


class TestPseudoAgentDBBindingWorker(base.DietTestCase):
    """Test class for AgentDBPortBinding."""

    def setUp(self):
        """Setup test."""
        self.useFixture(base.OpenDaylightRestClientFixture())
        self.useFixture(base.OpenDaylightPseudoAgentPrePopulateFixture())
        self.useFixture(OpenDaylightAgentDBFixture())
        super(TestPseudoAgentDBBindingWorker, self).setUp()

        self.worker = pseudo_agentdb_binding.PseudoAgentDBBindingWorker()

    def test_update_agents_db(self):
        """test agent update."""
        self.worker.update_agents_db(
            hostconfigs=SAMPLE_ODL_HCONFIGS['hostconfigs']['hostconfig'])
        self.worker.agents_db.create_or_update_agent.assert_called_once()


class TestPseudoAgentDBBindingController(base.DietTestCase):
    """Test class for AgentDBPortBinding."""

    # Test data for string interpolation of substitutable identifers
    #   e.g. $PORT_ID identifier in the configurations JSON string  below shall
    # be substituted with portcontext.current['id'] eliminating the check
    # for specific vif_type making port-binding truly switch agnostic.
    # Refer: Python string templates and interpolation (string.Template)
    sample_hconf_str_tmpl_subs_vpp = {
        "host": "devstack",      # host-id in ODL JSON
        "agent_type": "ODL L2",  # host-type in ODL JSON
                                 # config in ODL JSON
        "alive": True,
        "configurations": {"supported_vnic_types": [
            {"vnic_type": "normal", "vif_type": "vhostuser",
             "vif_details": {
                 "uuid": "TEST_UUID",
                 "has_datapath_type_netdev": True,
                 "support_vhost_user": True,
                 "port_prefix": "socket_",
                 "vhostuser_socket_dir": "/tmp",
                 "vhostuser_ovs_plug": True,
                 "vhostuser_mode": "server",
                 "vhostuser_socket":
                     "/tmp/socket_$PORT_ID"
             }}],
            "allowed_network_types": [
                "local", "vlan", "vxlan", "gre"],
            "bridge_mappings": {"physnet1": "br-ex"}}
    }

    sample_hconf_str_tmpl_subs_ovs = {
        "host": "devstack",      # host-id in ODL JSON
        "agent_type": "ODL L2",  # host-type in ODL JSON
                                 # config in ODL JSON
        "alive": True,
        "configurations": {"supported_vnic_types": [
            {"vnic_type": "normal", "vif_type": "vhostuser",
             "vif_details": {
                 "uuid": "TEST_UUID",
                 "has_datapath_type_netdev": True,
                 "support_vhost_user": True,
                 "port_prefix": "vhu",
                 "vhostuser_socket_dir": "/var/run/openvswitch",
                 "vhostuser_ovs_plug": True,
                 "vhostuser_mode": "client",
                 "vhostuser_socket":
                     "/var/run/openvswitch/vhu$PORT_ID"
             }}],
            "allowed_network_types": [
                "local", "vlan", "vxlan", "gre"],
            "bridge_mappings": {"physnet1": "br-ex"}}
    }

    sample_hconf_str_tmpl_nosubs = {
        "host": "devstack",      # host-id in ODL JSON
        "agent_type": "ODL L2",  # host-type in ODL JSON
                                 # config in ODL JSON
        "configurations": {"supported_vnic_types": [
            {"vnic_type": "normal", "vif_type": "ovs",
             "vif_details": {
                 "uuid": "TEST_UUID",
                 "has_datapath_type_netdev": True,
                 "support_vhost_user": True,
                 "port_prefix": "socket_",
                 "vhostuser_socket_dir": "/tmp",
                 "vhostuser_ovs_plug": True,
                 "vhostuser_mode": "server",
                 "vhostuser_socket":
                     "/var/run/openvswitch/PORT_NOSUBS"
             }}],
            "allowed_network_types": [
                "local", "vlan", "vxlan", "gre"],
            "bridge_mappings": {"physnet1": "br-ex"}}
    }

    # Test data for vanilla OVS
    sample_hconfig_dbget_ovs = {"configurations": {"supported_vnic_types": [
        {"vnic_type": "normal", "vif_type": portbindings.VIF_TYPE_OVS,
         "vif_details": {
             "some_test_details": None
         }}],
        "allowed_network_types": ["local", "vlan", "vxlan", "gre"],
        "bridge_mappings": {"physnet1": "br-ex"}}}

    # Test data for vanilla OVS with SR-IOV offload
    sample_hconfig_dbget_ovs_sriov_offload = {"configurations": {
        "supported_vnic_types": [{
            "vnic_type": "normal",
            "vif_type": portbindings.VIF_TYPE_OVS,
            "vif_details": {
                "some_test_details": None}}, {
            "vnic_type": "direct",
            "vif_type": portbindings.VIF_TYPE_OVS,
            "vif_details": {
                "some_test_details": None
            }}, ],
        "allowed_network_types": ["local", "vlan", "vxlan", "gre"],
        "bridge_mappings": {"physnet1": "br-ex"}}}

    # Test data for OVS-DPDK
    sample_hconfig_dbget_ovs_dpdk = {"configurations": {
        "supported_vnic_types": [{
            "vnic_type": "normal",
            "vif_type": portbindings.VIF_TYPE_VHOST_USER,
            "vif_details": {
                "uuid": "TEST_UUID",
                "has_datapath_type_netdev": True,
                "support_vhost_user": True,
                "port_prefix": "vhu",
                # Assumption: /var/run mounted as tmpfs
                "vhostuser_socket_dir": "/var/run/openvswitch",
                "vhostuser_ovs_plug": True,
                "vhostuser_mode": "client",
                "vhostuser_socket": "/var/run/openvswitch/vhu$PORT_ID"}}],
        "allowed_network_types": ["local", "vlan", "vxlan", "gre"],
        "bridge_mappings": {"physnet1": "br-ex"}}}

    # Test data for VPP
    sample_hconfig_dbget_vpp = {"configurations": {"supported_vnic_types": [
        {"vnic_type": "normal", "vif_type": portbindings.VIF_TYPE_VHOST_USER,
         "vif_details": {
             "uuid": "TEST_UUID",
             "has_datapath_type_netdev": True,
             "support_vhost_user": True,
             "port_prefix": "socket_",
             "vhostuser_socket_dir": "/tmp",
             "vhostuser_ovs_plug": True,
             "vhostuser_mode": "server",
             "vhostuser_socket": "/tmp/socket_$PORT_ID"
         }}],
        "allowed_network_types": ["local", "vlan", "vxlan", "gre"],
        "bridge_mappings": {"physnet1": "br-ex"}}}

    # Test data for length of string
    sample_odl_hconfigs_length = {
        "host": "devstack",      # host-id in ODL JSON
        "agent_type": "ODL L2",  # host-type in ODL JSON
                                 # config in ODL JSON
        "configurations": {"supported_vnic_types": [
            {"vnic_type": "normal", "vif_type": "vhostuser",
             "vif_details": {
                 "uuid": "TEST_UUID",
                 "has_datapath_type_netdev": True,
                 "support_vhost_user": True,
                 "port_prefix": "longprefix_",
                 "vhostuser_socket_dir": "/tmp",
                 "vhostuser_ovs_plug": True,
                 "vhostuser_mode": "server",
                 "vhostuser_socket":
                     "/tmp/longprefix_$PORT_ID"
             }}],
            "allowed_network_types": [
                "local", "vlan", "vxlan", "gre"],
            "bridge_mappings": {"physnet1": "br-ex"}}
    }

    # Raw test data for unicode/string comparison
    sample_odl_hconfigs_length_raw = {
        "host": "devstack",
        "agent_type": "ODL L2",
        "configurations": """{"supported_vnic_types": [
            {"vnic_type": "normal", "vif_type": "vhostuser",
             "vif_details": {
                 "uuid": "TEST_UUID",
                 "has_datapath_type_netdev": true,
                 "support_vhost_user": true,
                 "port_prefix": "prefix_",
                 "vhostuser_socket_dir": "/tmp",
                 "vhostuser_ovs_plug": true,
                 "vhostuser_mode": "server",
                 "vhostuser_socket":
                     "/tmp/prefix_$PORT_ID"
             }}],
            "allowed_network_types": [
                "local", "vlan", "vxlan", "gre"],
            "bridge_mappings": {"physnet1": "br-ex"}}"""
    }

    # test data valid  and invalid segments
    test_valid_segment = {
        api.ID: 'API_ID',
        api.NETWORK_TYPE: n_const.TYPE_LOCAL,
        api.SEGMENTATION_ID: 'API_SEGMENTATION_ID',
        api.PHYSICAL_NETWORK: 'API_PHYSICAL_NETWORK'}

    test_invalid_segment = {
        api.ID: 'API_ID',
        api.NETWORK_TYPE: n_const.TYPE_NONE,
        api.SEGMENTATION_ID: 'API_SEGMENTATION_ID',
        api.PHYSICAL_NETWORK: 'API_PHYSICAL_NETWORK'}

    def setUp(self):
        """Setup test."""
        self.useFixture(base.OpenDaylightRestClientFixture())
        self.useFixture(base.OpenDaylightFeaturesFixture())
        self.useFixture(base.OpenDaylightPseudoAgentPrePopulateFixture())
        self.useFixture(OpenDaylightAgentDBFixture())
        super(TestPseudoAgentDBBindingController, self).setUp()
        self.useFixture(fixture.CallbackRegistryFixture())
        self.cfg = self.useFixture(config_fixture.Config())

        self.mgr = pseudo_agentdb_binding.PseudoAgentDBBindingController()

    def test_is_valid_segment(self):
        """Validate the _check_segment method."""
        all_network_types = [n_const.TYPE_FLAT, n_const.TYPE_GRE,
                             n_const.TYPE_LOCAL, n_const.TYPE_VXLAN,
                             n_const.TYPE_VLAN, n_const.TYPE_NONE]

        valid_types = {
            network_type
            for network_type in all_network_types
            if self.mgr._is_valid_segment(
                {api.NETWORK_TYPE: network_type},
                {'allowed_network_types': [
                    n_const.TYPE_LOCAL, n_const.TYPE_GRE,
                    n_const.TYPE_VXLAN, n_const.TYPE_VLAN]})}

        self.assertEqual({
            n_const.TYPE_LOCAL, n_const.TYPE_GRE, n_const.TYPE_VXLAN,
            n_const.TYPE_VLAN}, valid_types)

    def test_bind_port_with_vif_type_ovs(self):
        """test bind_port with vanilla ovs."""
        port_context = self._fake_port_context(
            fake_segments=[self.test_invalid_segment, self.test_valid_segment])

        vif_type = portbindings.VIF_TYPE_OVS
        vif_details = {'some_test_details': None}

        self.mgr._hconfig_bind_port(
            port_context, self.sample_hconfig_dbget_ovs)

        port_context.set_binding.assert_called_once_with(
            self.test_valid_segment[api.ID], vif_type,
            vif_details, status=n_const.PORT_STATUS_ACTIVE)

    def test_bind_port_with_vif_type_ovs_with_sriov_offload(self):
        """test bind_port with vanilla ovs with SR-IOV offload"""
        port_context = self._fake_port_context(
            fake_segments=[self.test_invalid_segment, self.test_valid_segment])

        vif_type = portbindings.VIF_TYPE_OVS
        vif_details = {'some_test_details': None}

        self.mgr._hconfig_bind_port(
            port_context, self.sample_hconfig_dbget_ovs_sriov_offload)

        port_context.set_binding.assert_called_once_with(
            self.test_valid_segment[api.ID], vif_type,
            vif_details, status=n_const.PORT_STATUS_ACTIVE)

    def _set_pass_vif_details(self, port_context, vif_details):
        """extract vif_details and update vif_details if needed."""
        vhostuser_socket_dir = vif_details.get(
            'vhostuser_socket_dir', '/var/run/openvswitch')
        port_spec = vif_details.get(
            'port_prefix', 'vhu') + port_context.current['id']
        socket_path = os_path.join(vhostuser_socket_dir, port_spec)
        vif_details.update({portbindings.VHOST_USER_SOCKET: socket_path})

        return vif_details

    def test_bind_port_with_vif_type_vhost_user(self):
        """test bind_port with ovs-dpdk."""
        port_context = self._fake_port_context(
            fake_segments=[self.test_invalid_segment, self.test_valid_segment],
            host_agents=[deepcopy(self.sample_hconf_str_tmpl_subs_ovs)])

        self.mgr.bind_port(port_context)

        pass_vif_type = portbindings.VIF_TYPE_VHOST_USER
        pass_vif_details = self.sample_hconfig_dbget_ovs_dpdk[
            'configurations']['supported_vnic_types'][0]['vif_details']
        self._set_pass_vif_details(port_context, pass_vif_details)

        port_context.set_binding.assert_called_once_with(
            self.test_valid_segment[api.ID], pass_vif_type,
            pass_vif_details, status=n_const.PORT_STATUS_ACTIVE)

    def _test_bind_port_succeed_when_agent_status(self, hconfig, agent_status):
        hconfig['alive'] = agent_status
        port_context = self._fake_port_context(
            fake_segments=[self.test_invalid_segment, self.test_valid_segment],
            host_agents=[hconfig])
        self.mgr.bind_port(port_context)
        port_context.set_binding.assert_called()

    def test_bind_port_succeed_when_agent_dead_vpp(self):
        hconfig = deepcopy(self.sample_hconf_str_tmpl_subs_vpp)
        self._test_bind_port_succeed_when_agent_status(hconfig, False)

    def test_bind_port_succeed_when_agent_dead_ovs(self):
        hconfig = deepcopy(self.sample_hconf_str_tmpl_subs_ovs)
        self._test_bind_port_succeed_when_agent_status(hconfig, False)

    def test_bind_port_succeed_when_agent_alive_vpp(self):
        hconfig = deepcopy(self.sample_hconf_str_tmpl_subs_vpp)
        self._test_bind_port_succeed_when_agent_status(hconfig, True)

    def test_bind_port_succeed_when_agent_alive_ovs(self):
        hconfig = deepcopy(self.sample_hconf_str_tmpl_subs_ovs)
        self._test_bind_port_succeed_when_agent_status(hconfig, True)

    def test_bind_port_with_vif_type_vhost_user_vpp(self):
        """test bind_port with vpp."""
        port_context = self._fake_port_context(
            fake_segments=[self.test_invalid_segment, self.test_valid_segment],
            host_agents=[deepcopy(self.sample_hconf_str_tmpl_subs_vpp)])

        self.mgr.bind_port(port_context)

        pass_vif_type = portbindings.VIF_TYPE_VHOST_USER
        pass_vif_details = self.sample_hconfig_dbget_vpp['configurations'][
            'supported_vnic_types'][0]['vif_details']
        self._set_pass_vif_details(port_context, pass_vif_details)

        port_context.set_binding.assert_called_once_with(
            self.test_valid_segment[api.ID], pass_vif_type,
            pass_vif_details, status=n_const.PORT_STATUS_ACTIVE)

    def test_bind_port_without_valid_segment(self):
        """test bind_port without a valid segment."""
        port_context = self._fake_port_context(
            fake_segments=[self.test_invalid_segment])

        self.mgr._hconfig_bind_port(
            port_context, self.sample_hconfig_dbget_ovs)

        port_context.set_binding.assert_not_called()

    def test_no_str_template_substitution_in_configuration_string(self):
        """Test for no identifier substituion in config JSON string."""
        port_context = self._fake_port_context(
            fake_segments=[self.test_invalid_segment, self.test_valid_segment])

        hconf_dict = self.mgr._substitute_hconfig_tmpl(
            port_context, self.sample_hconf_str_tmpl_nosubs)

        test_string = hconf_dict['configurations'][
            'supported_vnic_types'][0][
                'vif_details'][portbindings.VHOST_USER_SOCKET]

        expected_str = '/var/run/openvswitch/PORT_NOSUBS'

        self.assertEqual(expected_str, test_string)

    def test_str_template_substitution_in_configuration_string(self):
        """Test for identifier substitution in config JSON string."""
        port_context = self._fake_port_context(
            fake_segments=[self.test_invalid_segment, self.test_valid_segment])

        hconf_dict = self.mgr._substitute_hconfig_tmpl(
            port_context, self.sample_hconf_str_tmpl_subs_vpp)

        test_string = hconf_dict['configurations'][
            'supported_vnic_types'][0][
                'vif_details'][portbindings.VHOST_USER_SOCKET]

        expected_str = Template('/tmp/socket_$PORT_ID')
        expected_str = expected_str.safe_substitute({
            'PORT_ID': port_context.current['id']})

        self.assertEqual(expected_str, test_string)

    def test_str_template_substitution_length_in_configuration_string(self):
        """Test for identifier substitution in config JSON string."""
        port_context = self._fake_port_context(
            fake_segments=[self.test_invalid_segment, self.test_valid_segment])

        hconf_dict = self.mgr._substitute_hconfig_tmpl(
            port_context, self.sample_odl_hconfigs_length)

        test_string = hconf_dict['configurations'][
            'supported_vnic_types'][0][
            'vif_details'][portbindings.VHOST_USER_SOCKET]

        expected_str = Template('/tmp/longprefix_$PORT_ID')
        expected_str = expected_str.safe_substitute({
            'PORT_ID': port_context.current['id']})

        self.assertNotEqual(expected_str, test_string)
        self.assertEqual(len(test_string) - len('/tmp/'), 14)

    def test_template_substitution_in_raw_configuration(self):
        """Test for identifier substitution in config string."""
        port_context = self._fake_port_context(
            fake_segments=[self.test_invalid_segment, self.test_valid_segment])

        # Substitute raw string configuration with json
        raw_configurations = self.sample_odl_hconfigs_length_raw[
            'configurations']
        raw_configurations_json = jsonutils.loads(raw_configurations)
        self.sample_odl_hconfigs_length_raw['configurations'] = (
            raw_configurations_json)

        hconf_dict = self.mgr._substitute_hconfig_tmpl(
            port_context, self.sample_odl_hconfigs_length_raw)

        test_string = hconf_dict['configurations'][
            'supported_vnic_types'][0][
                'vif_details'][portbindings.VHOST_USER_SOCKET]

        expected_str = Template('/tmp/prefix_$PORT_ID')
        expected_str = expected_str.safe_substitute({
            'PORT_ID': port_context.current['id']})

        self.assertEqual(expected_str, test_string)

    def _fake_port_context(self, fake_segments, host_agents=None):
        network = mock.MagicMock(spec=api.NetworkContext)
        return mock.MagicMock(
            spec=ctx.PortContext,
            current={'id': 'PORTID',
                     portbindings.VNIC_TYPE: portbindings.VNIC_NORMAL},
            segments_to_bind=fake_segments, network=network,
            host_agents=lambda agent_type: host_agents,
            _plugin_context=mock.MagicMock()
        )

    @mock.patch.object(provisioning_blocks, 'add_provisioning_component')
    def test_prepare_inital_port_status_no_websocket(
            self, mocked_add_provisioning_component):
        port_ctx = self._fake_port_context(
            fake_segments=[self.test_valid_segment])
        initial_port_status = self.mgr._prepare_initial_port_status(port_ctx)
        self.assertEqual(initial_port_status, n_const.PORT_STATUS_ACTIVE)
        mocked_add_provisioning_component.assert_not_called()

    @mock.patch.object(provisioning_blocks, 'add_provisioning_component')
    def test_prepare_inital_port_status_with_websocket(
            self, mocked_add_provisioning_component):
        feature_json = """{"features": {"feature":
                            [{"service-provider-feature":
                            "neutron-extensions:operational-port-status"}]}}"""
        self.cfg.config(odl_features_json=feature_json, group='ml2_odl')
        self.addCleanup(odl_features.deinit)
        odl_features.init()
        port_ctx = self._fake_port_context(
            fake_segments=[self.test_valid_segment])
        initial_port_status = self.mgr._prepare_initial_port_status(port_ctx)
        self.assertEqual(initial_port_status, n_const.PORT_STATUS_DOWN)
        mocked_add_provisioning_component.assert_called()


class TestPseudoAgentDBBindingControllerBug1608659(
        test_plugin.NeutronDbPluginV2TestCase):
    """Test class for Bug1608659."""

    # test data hostconfig
    sample_odl_hconfigs = {"hostconfigs": {"hostconfig": [
        {"host-id": "devstack-control",
         "host-type": "ODL L2",
         "config": """{"supported_vnic_types": [
             {"vnic_type": "normal", "vif_type": "vhostuser",
              "vif_details":
                  {"port_filter": "False",
                   "vhostuser_socket": "/var/run/openvswitch"}}],
             "allowed_network_types": [
                 "local", "vlan", "vxlan", "gre"],
             "bridge_mappings": {"physnet1": "br-ex"}}"""},
        {"host-id": "devstack-control",
         "host-type": "ODL L3",
         "config": """{ "some_details": "dummy_details" }"""}
    ]}}

    def setUp(self):
        self.useFixture(base.OpenDaylightRestClientFixture())
        self.useFixture(base.OpenDaylightPseudoAgentPrePopulateFixture())
        self.useFixture(OpenDaylightAgentDBFixture())
        super(TestPseudoAgentDBBindingControllerBug1608659, self).setUp(
            plugin='ml2')
        self.worker = pseudo_agentdb_binding.PseudoAgentDBBindingWorker()

    def test_execute_no_exception(self):
        with mock.patch.object(pseudo_agentdb_binding, 'LOG') as mock_log:
            self.worker.update_agents_db(
                self.sample_odl_hconfigs['hostconfigs']['hostconfig'])
            # Assert no exception happened
            self.assertFalse(mock_log.exception.called)


class TestPseudoAgentNeutronWorker(testlib_api.SqlTestCase):
    def setUp(self):
        self.useFixture(base.OpenDaylightRestClientFixture())
        self.useFixture(base.OpenDaylightJournalThreadFixture())
        self.useFixture(base.OpenDaylightFeaturesFixture())
        self.useFixture(base.OpenDaylightPseudoAgentPrePopulateFixture())
        self.cfg = self.useFixture(config_fixture.Config())
        self.mock_periodic_thread = mock.patch.object(
            periodic_task.PeriodicTask, 'start').start()
        super(TestPseudoAgentNeutronWorker, self).setUp()
        self.cfg.config(mechanism_drivers=['opendaylight_v2'], group='ml2')
        self.cfg.config(
            port_binding_controller='pseudo-agentdb-binding', group='ml2_odl')

    def test_get_worker(self):
        workers = ml2_plugin.Ml2Plugin().get_workers()
        self.assertTrue(any(
            isinstance(worker,
                       pseudo_agentdb_binding.PseudoAgentDBBindingWorker)
            for worker in workers))

    def test_worker(self):
        worker = pseudo_agentdb_binding.PseudoAgentDBBindingWorker()
        worker.wait()
        worker.stop()
        worker.reset()

    def test_worker_start_websocket(self):
        self.cfg.config(enable_websocket_pseudo_agentdb=True, group='ml2_odl')
        worker = pseudo_agentdb_binding.PseudoAgentDBBindingWorker()
        with mock.patch.object(
                websocket_client.OpenDaylightWebsocketClient,
                'odl_create_websocket') as mock_odl_create_websocket:
            worker.start()
            mock_odl_create_websocket.assert_called_once()

    def test_worker_start_periodic(self):
        self.cfg.config(enable_websocket_pseudo_agentdb=False, group='ml2_odl')
        worker = pseudo_agentdb_binding.PseudoAgentDBBindingWorker()
        with mock.patch.object(
                periodic_task.PeriodicTask, 'start') as mock_start:
            worker.start()
            mock_start.assert_called_once()



# Copyright (c) 2017 Red Hat, Inc.
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import contextlib
import io
import random
import string
from unittest import mock

import fixtures

from networking_odl.cmd import analyze_journal
from networking_odl.journal import journal
from networking_odl.tests import base


def _random_string():
    letters = string.ascii_lowercase
    return ''.join(random.choice(letters)
                   for _ in range(random.randint(1, 10)))


def _generate_log_entry(log_type=None, entry_id=None):
    entry = mock.Mock()
    entry.seqnum = entry_id if entry_id else _random_string()
    entry.operation = _random_string()
    entry.object_type = _random_string()
    entry.object_uuid = _random_string()

    logger = fixtures.FakeLogger()
    with logger:
        journal._log_entry(log_type if log_type else _random_string(), entry)

    return entry, "noise %s noise" % logger.output


class TestAnalyzeJournal(base.DietTestCase):
    def setUp(self):
        super(TestAnalyzeJournal, self).setUp()
        self.output = io.StringIO()

    def _assert_nothing_printed(self):
        self.assertEqual('', self.output.getvalue())

    def _assert_something_printed(self, expected=None):
        self.assertNotEqual('', self.output.getvalue())
        if expected:
            self.assertIn(str(expected), self.output.getvalue())

    def test_setup_conf_no_args(self):
        conf = analyze_journal.setup_conf(self.output, [])
        self.assertIsNotNone(conf)
        self._assert_nothing_printed()

    def test_setup_conf_h_flag(self):
        self.assertRaises(
            SystemExit, analyze_journal.setup_conf, self.output, ['-h'])
        self._assert_something_printed()

    def test_setup_conf_help_flag(self):
        self.assertRaises(
            SystemExit, analyze_journal.setup_conf, self.output, ['--help'])
        self._assert_something_printed()

    def test_setup_conf_file(self):
        file_name = _random_string()
        conf = analyze_journal.setup_conf(self.output, ['--file', file_name])
        self.assertEqual(file_name, conf.file)

    def test_setup_conf_slowest(self):
        slowest = random.randint(1, 10000)
        conf = analyze_journal.setup_conf(
            self.output, ['--slowest', str(slowest)])
        self.assertEqual(slowest, conf.slowest)

    def test_setup_conf_slowest_zero(self):
        self.assertRaises(SystemExit, analyze_journal.setup_conf,
                          self.output, ['--slowest', '0'])
        self._assert_nothing_printed()

    def test_parse_log_no_matched_content(self):
        self.assertEqual({}, analyze_journal.parse_log([]))
        self.assertEqual({}, analyze_journal.parse_log(['dummy']))

    def _test_parse_log_entry(self, recorded=False, completed=False):
        content = []
        entry_id = _random_string()
        entry = None

        if recorded:
            entry, log = _generate_log_entry(log_type=journal.LOG_RECORDED,
                                             entry_id=entry_id)
            content.append(log)

        if completed:
            centry, log = _generate_log_entry(log_type=journal.LOG_COMPLETED,
                                              entry_id=entry_id)
            entry = centry if entry is None else entry
            content.append(log)

        entries = analyze_journal.parse_log(content)
        actual_entry = entries[entry_id]
        self.assertEqual(entry.operation, actual_entry['op'])
        self.assertEqual(entry.object_type, actual_entry['obj_type'])
        self.assertEqual(entry.object_uuid, actual_entry['obj_id'])

        if recorded:
            self.assertGreater(actual_entry[journal.LOG_RECORDED], 0)

        if completed:
            self.assertGreater(actual_entry[journal.LOG_COMPLETED], 0)

    def test_parse_log_entry_recorded(self):
        self._test_parse_log_entry(recorded=True)

    def test_parse_log_entry_completed(self):
        self._test_parse_log_entry(completed=True)

    def test_parse_log_entry_recorded_and_completed(self):
        self._test_parse_log_entry(recorded=True, completed=True)

    def test_analyze_entries_no_records(self):
        self.assertEqual([], analyze_journal.analyze_entries({}))

    def _generate_random_entry(self):
        return dict([(k, _random_string()) for k in analyze_journal.LOG_KEYS])

    def _entry_for_analyze_entries(self, recorded=False, completed=False):
        entry = self._generate_random_entry()

        if recorded:
            entry[journal.LOG_RECORDED] = random.uniform(1, 10)

        if completed:
            entry[journal.LOG_COMPLETED] = random.uniform(10, 20)

        return entry

    def test_analyze_entries_no_completed_time(self):
        entry = self._entry_for_analyze_entries(recorded=True)
        entries = {entry['entry_id']: entry}
        self.assertEqual([], analyze_journal.analyze_entries(entries))

    def test_analyze_entries_no_recorded_time(self):
        entry = self._entry_for_analyze_entries(completed=True)
        entries = {entry['entry_id']: entry}
        self.assertEqual([], analyze_journal.analyze_entries(entries))

    def test_analyze_entries(self):
        entry = self._entry_for_analyze_entries(recorded=True, completed=True)
        entry_only_recorded = self._entry_for_analyze_entries(recorded=True)
        entry_only_completed = self._entry_for_analyze_entries(completed=True)

        entries = {e['entry_id']: e for e in
                   (entry, entry_only_recorded, entry_only_completed)}
        entries_stats = analyze_journal.analyze_entries(entries)

        expected_time = (entry[journal.LOG_COMPLETED] -
                         entry[journal.LOG_RECORDED])
        expected_entry = analyze_journal.EntryStats(
            entry_id=entry['entry_id'], time=expected_time, op=entry['op'],
            obj_type=entry['obj_type'], obj_id=entry['obj_id'])
        self.assertIn(expected_entry, entries_stats)

    def _assert_percentile_printed(self, entries_stats, percentile):
        expected_percentile_format = "%sth percentile: %s"
        percentile_index = int(len(entries_stats) * (percentile / 100.0))
        entry = entries_stats[percentile_index]

        self._assert_something_printed(expected_percentile_format %
                                       (percentile, entry.time))

    def test_print_stats(self):
        entries_stats = []
        entries_count = 10
        slowest = random.randint(1, int(entries_count / 2))
        for i in range(entries_count):
            entry = self._generate_random_entry()
            entries_stats.append(
                analyze_journal.EntryStats(
                    entry_id=entry['entry_id'], time=i, op=entry['op'],
                    obj_type=entry['obj_type'], obj_id=entry['obj_id']))

        analyze_journal.print_stats(self.output, slowest, entries_stats)

        total_time = (entries_count * (entries_count - 1)) / 2
        avg_time = total_time / entries_count
        self._assert_something_printed(avg_time)

        self._assert_something_printed(slowest)
        self._assert_percentile_printed(entries_stats, 90)
        self._assert_percentile_printed(entries_stats, 99)
        self._assert_percentile_printed(entries_stats, 99.9)

        expected = ''
        for i in reversed(range(entries_count - slowest, entries_count)):
            entry = entries_stats[i]
            expected += '\n'
            expected += (analyze_journal.ENTRY_LOG_TEMPLATE %
                         (entry.entry_id, entry.time, entry.op, entry.obj_type,
                          entry.obj_id))
            self._assert_something_printed(expected)

    @contextlib.contextmanager
    def _setup_mocks_for_main(self, content):
        with mock.patch.object(analyze_journal, 'get_content') as mgc, \
                mock.patch.object(analyze_journal, 'setup_conf') as msc:
            m = mock.MagicMock()
            m.__iter__.return_value = content
            mgc().__enter__.return_value = m
            conf = msc()
            conf.slowest = 10
            yield

    def test_main(self):
        entry_id = _random_string()
        _, entry_recorded = _generate_log_entry(journal.LOG_RECORDED, entry_id)
        _, entry_completed = _generate_log_entry(journal.LOG_COMPLETED,
                                                 entry_id)
        with self._setup_mocks_for_main((entry_recorded, entry_completed)):
            rc = analyze_journal.main(self.output)

        self.assertEqual(0, rc)
        self._assert_something_printed(entry_id)

    def test_main_no_entry_stats(self):
        with self._setup_mocks_for_main(('dummy',)):
            rc = analyze_journal.main(self.output)

        self.assertNotEqual(0, rc)
        self._assert_something_printed()



# Copyright (c) 2016 OpenStack Foundation
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

# pylint: disable=unused-argument, protected-access


from contextlib import contextmanager
import io
import os
import sys
import tempfile
from unittest import mock

from oslo_serialization import jsonutils

from networking_odl.cmd import set_ovs_hostconfigs
from networking_odl.tests import base
from networking_odl.tests import match

LOGGING_ENABLED = "Logging Enabled!"
LOGGING_PERMISSION_REQUIRED = "permissions are required to configure ovsdb"


@contextmanager
def capture(command, args):
    out, sys.stdout = sys.stdout, io.StringIO()
    try:
        command(args)
        sys.stdout.seek(0)
        yield sys.stdout.read()
    finally:
        sys.stdout = out


class TestSetOvsHostconfigs(base.DietTestCase):

    maxDiff = None

    def test_given_ovs_hostconfigs(self):
        # given
        self.patch_os_geteuid()
        ovs_hostconfigs = {
            "ODL L2": {"allowed_network_types": ["a", "b", "c"]}}
        args = ['--ovs_hostconfigs=' + jsonutils.dumps(ovs_hostconfigs),
                '--bridge_mappings=a:1,b:2']
        execute = self.patch_utils_execute()
        conf = set_ovs_hostconfigs.setup_conf(args)

        # when
        result = set_ovs_hostconfigs.main(args)

        # then
        self.assertEqual(0, result)
        execute.assert_has_calls([
            mock.call(
                ('ovs-vsctl', 'get', 'Open_vSwitch', '.', '_uuid')),
            mock.call(
                ('ovs-vsctl', 'set', 'Open_vSwitch', '<some-uuid>',
                 'external_ids:odl_os_hostconfig_hostid=' + conf.host)),
            mock.call(
                ('ovs-vsctl', 'set', 'Open_vSwitch', '<some-uuid>',
                 match.wildcard(
                     'external_ids:odl_os_hostconfig_config_odl_l2=*'))),
        ])

        expected = ovs_hostconfigs['ODL L2']
        _, actual_json = execute.call_args_list[2][0][0][4].split("=", 1)
        self.assertEqual(match.json(expected), actual_json)

    def test_given_no_args(self):
        self._test_given_args(tuple())

    def test_given_default_values(self):
        self._test_given_args([])

    def test_given_datapath_type_system(self):
        self._test_given_args(['--datapath_type=netdev'])

    def test_given_datapath_type_netdev(self):
        self._test_given_args(['--datapath_type=netdev'])

    def test_given_datapath_type_vhostuser(self):
        self._test_given_args(['--datapath_type=dpdkvhostuser'])

    def test_given_ovs_dpdk(self):
        self._test_given_args(['--ovs_dpdk'])

    def test_given_noovs_dpdk(self):
        self._test_given_args(['--noovs_dpdk'])

    def test_given_ovs_sriov_offload(self):
        self._test_given_args(['--noovs_dpdk', '--ovs_sriov_offload'])

    def test_given_vhostuser_ovs_plug(self):
        self._test_given_args(['--vhostuser_ovs_plug'])

    def test_given_novhostuser_ovs_plug(self):
        self._test_given_args(['--novhostuser_ovs_plug'])

    def test_given_allowed_network_types(self):
        self._test_given_args(['--allowed_network_types=a,b,c'])

    def test_given_local_ip(self):
        self._test_given_args(['--local_ip=192.168.1.10', '--host='])

    def test_given_vhostuser_mode_server(self):
        self._test_given_args(
            ['--vhostuser_mode=server', '--datapath_type=netdev'])

    def test_given_vhostuser_mode_client(self):
        self._test_given_args(
            ['--vhostuser_mode=client', '--datapath_type=netdev'])

    def test_given_vhostuser_port_prefix_vhu(self):
        self._test_given_args(
            ['--vhostuser_port_prefix=vhu', '--datapath_type=netdev'])

    def test_given_vhostuser_port_prefix_socket(self):
        self._test_given_args(
            ['--vhostuser_port_prefix=socket', '--datapath_type=netdev'])

    def test_given_config_file(self):
        file_descriptor, file_path = tempfile.mkstemp()

        try:
            os.write(file_descriptor, b"# dummy neutron config file\n")
            os.close(file_descriptor)
            self._test_given_args(['--config-file={}'.format(file_path)])

        finally:
            os.remove(file_path)

    def _test_given_args(self, *args):
        # given
        self.patch_os_geteuid()
        execute = self.patch_utils_execute()
        conf = set_ovs_hostconfigs.setup_conf(*args)

        datapath_type = conf.datapath_type
        if datapath_type is None:
            if conf.ovs_dpdk is False:
                datapath_type = "system"
            else:
                datapath_type = "netdev"

        # when
        result = set_ovs_hostconfigs.main(*args)

        # then
        self.assertEqual(0, result)
        execute.assert_has_calls([
            mock.call(
                ('ovs-vsctl', 'get', 'Open_vSwitch', '.', '_uuid')),
            mock.call(
                ('ovs-vsctl', 'get', 'Open_vSwitch', '.', 'datapath_types')),
            mock.call(
                ('ovs-vsctl', 'set', 'Open_vSwitch', '<some-uuid>',
                 'external_ids:odl_os_hostconfig_hostid=' + conf.host)),
            mock.call(
                ('ovs-vsctl', 'set', 'Open_vSwitch', '<some-uuid>',
                 match.wildcard(
                     'external_ids:odl_os_hostconfig_config_odl_l2=*'))),
        ])

        host_addresses = [conf.host or conf.local_ip]
        if datapath_type == "system":
            vif_type = "ovs"
            vif_details = {
                "uuid": '<some-uuid>',
                "host_addresses": host_addresses,
                "has_datapath_type_netdev": False,
                "support_vhost_user": False
            }
        else:  # datapath_type in ["system", "netdev"]
            vif_type = "vhostuser"
            vif_details = {
                "uuid": '<some-uuid>',
                "host_addresses": host_addresses,
                "has_datapath_type_netdev": True,
                "support_vhost_user": True,
                "port_prefix": conf.vhostuser_port_prefix,
                "vhostuser_mode": conf.vhostuser_mode,
                "vhostuser_ovs_plug": conf.vhostuser_ovs_plug,
                "vhostuser_socket_dir": conf.vhostuser_socket_dir,
                "vhostuser_socket": os.path.join(
                    conf.vhostuser_socket_dir,
                    conf.vhostuser_port_prefix + "$PORT_ID"),
            }

        _, actual_json = execute.call_args_list[3][0][0][4].split("=", 1)
        expected = {
            "allowed_network_types": conf.allowed_network_types,
            "bridge_mappings": conf.bridge_mappings,
            "datapath_type": datapath_type,
            "supported_vnic_types": [
                {
                    "vif_type": vif_type,
                    "vnic_type": "normal",
                    "vif_details": vif_details
                }
            ]
        }

        if vif_type == 'ovs' and conf.ovs_sriov_offload:
            direct_vnic = {
                "vif_details": vif_details,
                "vif_type": vif_type,
                "vnic_type": "direct",
            }
            expected["supported_vnic_types"].append(direct_vnic)
        self.assertEqual(match.json(expected), actual_json)

    def test_given_ovs_dpdk_undetected(self):
        # given
        LOG = self.patch(set_ovs_hostconfigs, 'LOG')
        args = ('--ovs_dpdk', '--bridge_mappings=a:1,b:2', '--debug')
        conf = set_ovs_hostconfigs.setup_conf(args)
        self.patch_os_geteuid()
        execute = self.patch_utils_execute(datapath_types="whatever")

        # when
        result = set_ovs_hostconfigs.main(args)

        # then
        self.assertEqual(1, result)
        execute.assert_has_calls([
            mock.call(
                ('ovs-vsctl', 'get', 'Open_vSwitch', '.', '_uuid')),
            mock.call(
                ('ovs-vsctl', 'get', 'Open_vSwitch', '.', 'datapath_types')),
        ])
        LOG.error.assert_called_once_with(
            "Fatal error: %s",
            match.wildcard(
                "--ovs_dpdk option was specified but the 'netdev' "
                "datapath_type was not enabled. To override use option "
                "--datapath_type=netdev"), exc_info=conf.debug)

    def test_bridge_mappings(self):
        # when
        conf = set_ovs_hostconfigs.setup_conf(('--bridge_mappings=a:1,b:2',))
        self.assertEqual({'a': '1', 'b': '2'}, conf.bridge_mappings)

    def test_allowed_network_types(self):
        # when
        conf = set_ovs_hostconfigs.setup_conf(('--allowed_network_types=a,b',))
        self.assertEqual(['a', 'b'], conf.allowed_network_types)

    def patch_utils_execute(
            self, uuid='<some-uuid>',
            datapath_types='netdev,dpdkvhostuser,system'):

        def execute(args):
            command, method, table, record, value = args
            self.assertEqual('ovs-vsctl', command)
            self.assertEqual('Open_vSwitch', table)
            self.assertIn(method, ['get', 'set'])
            if method == 'set':
                self.assertEqual(uuid, record)
                return ""
            elif method == 'get':
                self.assertEqual('.', record)
                self.assertIn(value, ['_uuid', 'datapath_types'])
                if value == '_uuid':
                    return uuid
                elif value == 'datapath_types':
                    return datapath_types

            self.fail('Unexpected command: ' + repr(args))

        return self.patch(
            set_ovs_hostconfigs.subprocess, "check_output",
            side_effect=execute)

    def patch_os_geteuid(self, return_value=0):
        return self.patch(
            set_ovs_hostconfigs.os, "geteuid", return_value=return_value)

    @contextmanager
    def test_log_on_console_msg(self):
        with capture(set_ovs_hostconfigs.main, args=()) as output:
            self.assertNotEqual(-1, output.find(LOGGING_PERMISSION_REQUIRED))

    def test_log_in_file(self):
        with tempfile.TemporaryFile() as fp:
            set_ovs_hostconfigs.main(("--log-file=%s" % fp.name,))
            logs = [LOGGING_ENABLED, LOGGING_PERMISSION_REQUIRED]
            for line, count in fp.readline():
                self.assertNotEqual(-1, line.find(logs[count]))



# Copyright (c) 2016 OpenStack Foundation
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from unittest import mock

from neutron.tests.unit.plugins.ml2 import test_plugin

from networking_odl.common import client
from networking_odl.journal import base_driver
from networking_odl.journal import journal
from networking_odl.ml2 import mech_driver_v2
from networking_odl.tests import base
from networking_odl.tests.unit import test_base_db


class OpenDaylightConfigBase(test_plugin.Ml2PluginV2TestCase,
                             test_base_db.ODLBaseDbTestCase):
    def setUp(self):
        self.journal_thread_fixture = self.useFixture(
            base.OpenDaylightJournalThreadFixture())
        self.useFixture(base.OpenDaylightRestClientFixture())
        self.useFixture(base.OpenDaylightFullSyncFixture())
        super(OpenDaylightConfigBase, self).setUp()
        self.thread = journal.OpenDaylightJournalThread()
        self.addCleanup(base_driver.ALL_RESOURCES.clear)

    def run_journal_processing(self):
        """Cause the journal to process the first pending entry"""
        self.thread.sync_pending_entries()


class OpenDaylightTestCase(OpenDaylightConfigBase):
    def setUp(self):
        self.mock_sendjson = mock.patch.object(client.OpenDaylightRestClient,
                                               'sendjson').start()
        super(OpenDaylightTestCase, self).setUp()
        self.port_create_status = 'DOWN'
        self.mech = mech_driver_v2.OpenDaylightMechanismDriver()
        self.mock_sendjson.side_effect = self.check_sendjson

    def check_sendjson(self, method, urlpath, obj):
        self.assertFalse(urlpath.startswith("http://"))



# Copyright (c) 2017 Ericsson India Global Service Pvt Ltd.
# All Rights Reserved.
#
#  Licensed under the Apache License, Version 2.0 (the "License"); you may
#  not use this file except in compliance with the License. You may obtain
#  a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#  WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#  License for the specific language governing permissions and limitations
#  under the License.


import contextlib

from networking_odl.common import constants as odl_const
from networking_odl.tests.functional import base
from neutron.services.trunk import plugin as trunk_plugin
from neutron.tests.unit.plugins.ml2 import test_plugin
from neutron_lib.plugins import utils
from neutron_lib.services.trunk import constants

from oslo_utils import uuidutils


class _TrunkDriverTest(base.OdlTestsBase):

    def test_trunk_create(self):
        with self.trunk() as trunk:
            self.assert_resource_created(odl_const.ODL_TRUNK, trunk)

    def test_trunk_update(self):
        with self.trunk() as trunk:
            trunk['trunk'].update(admin_state_up=False)
            self.trunk_plugin.update_trunk(self.context,
                                           trunk['trunk']['id'], trunk)
            response = self.get_odl_resource(odl_const.ODL_TRUNK, trunk)
            self.assertFalse(response['trunk']['admin_state_up'])

    def test_subport_create(self):
        with self.trunk() as trunk:
            with self.subport() as subport:
                trunk_obj = self.trunk_plugin.add_subports(
                    self.context, trunk['trunk']['id'],
                    {'sub_ports': [subport]})
                response = self.get_odl_resource(odl_const.ODL_TRUNK,
                                                 {'trunk': trunk_obj})
                self.assertEqual(response['trunk']['sub_ports'][0]['port_id'],
                                 subport['port_id'])

    def test_subport_delete(self):
        with self.subport() as subport:
            with self.trunk([subport]) as trunk:
                response = self.get_odl_resource(odl_const.ODL_TRUNK, trunk)
                self.assertEqual(response['trunk']['sub_ports'][0]['port_id'],
                                 subport['port_id'])
                trunk_obj = self.trunk_plugin.remove_subports(
                    self.context, trunk['trunk']['id'],
                    {'sub_ports': [subport]})
                response = self.get_odl_resource(odl_const.ODL_TRUNK,
                                                 {'trunk': trunk_obj})
                self.assertEqual(response['trunk']['sub_ports'], [])

    def test_trunk_delete(self):
        with self.trunk() as trunk:
            self.trunk_plugin.delete_trunk(self.context, trunk['trunk']['id'])
            self.assertIsNone(self.get_odl_resource(odl_const.ODL_TRUNK,
                                                    trunk))

    @contextlib.contextmanager
    def trunk(self, subports=None):
        subports = subports if subports else []
        with self.network() as network:
            with self.subnet(network=network) as subnet:
                with self.port(subnet=subnet) as trunk_parent:
                    tenant_id = uuidutils.generate_uuid()
                    trunk = {'port_id': trunk_parent['port']['id'],
                             'tenant_id': tenant_id, 'project_id': tenant_id,
                             'admin_state_up': True,
                             'name': 'test_trunk', 'sub_ports': subports}
                    trunk_obj = self.trunk_plugin.create_trunk(
                        self.context, {'trunk': trunk})
                    yield {'trunk': trunk_obj}

    @contextlib.contextmanager
    def subport(self):
        with self.port() as child_port:
            subport = {'segmentation_type': 'vlan',
                       'segmentation_id': 123,
                       'port_id': child_port['port']['id']}
            yield subport


class TestTrunkV2Driver(base.V2DriverAdjustment, _TrunkDriverTest,
                        test_plugin.Ml2PluginV2TestCase):
    _mechanism_drivers = ['opendaylight_v2']

    def setUp(self):
        super(TestTrunkV2Driver, self).setUp()
        self.trunk_plugin = trunk_plugin.TrunkPlugin()
        self.trunk_plugin.add_segmentation_type(
            constants.SEGMENTATION_TYPE_VLAN, utils.is_valid_vlan_tag)



# Copyright (c) 2015 OpenStack Foundation
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import sqlalchemy as sa

from neutron_lib.db import model_base

from networking_odl.common import constants as odl_const


IdType = sa.BigInteger().with_variant(sa.Integer(), 'sqlite')

journal_dependencies = sa.Table(
    'opendaylight_journal_deps', model_base.BASEV2.metadata,
    sa.Column('depends_on', IdType,
              sa.ForeignKey('opendaylightjournal.seqnum', ondelete='CASCADE'),
              primary_key=True),
    sa.Column('dependent', IdType,
              sa.ForeignKey('opendaylightjournal.seqnum', ondelete='CASCADE'),
              primary_key=True))


class OpenDaylightJournal(model_base.BASEV2):
    __tablename__ = 'opendaylightjournal'

    seqnum = sa.Column(IdType, primary_key=True, autoincrement=True)
    object_type = sa.Column(sa.String(36), nullable=False)
    object_uuid = sa.Column(sa.String(36), nullable=False)
    operation = sa.Column(sa.String(36), nullable=False)
    data = sa.Column(sa.PickleType, nullable=True)
    state = sa.Column(sa.Enum(odl_const.PENDING, odl_const.PROCESSING,
                              odl_const.FAILED, odl_const.COMPLETED,
                              name='state'),
                      nullable=False, default=odl_const.PENDING)
    retry_count = sa.Column(sa.Integer, default=0)
    last_retried = sa.Column(sa.TIMESTAMP, server_default=sa.func.now(),
                             onupdate=sa.func.now())
    version_id = sa.Column(sa.Integer, server_default='0', nullable=False)
    dependencies = sa.orm.relationship(
        "OpenDaylightJournal", secondary=journal_dependencies,
        primaryjoin=seqnum == journal_dependencies.c.depends_on,
        secondaryjoin=seqnum == journal_dependencies.c.dependent,
        backref="depending_on"
    )

    __mapper_args__ = {
        'version_id_col': version_id
    }


class OpenDaylightPeriodicTask(model_base.BASEV2):
    __tablename__ = 'opendaylight_periodic_task'

    state = sa.Column(sa.Enum(odl_const.PENDING, odl_const.PROCESSING,
                              name='state'),
                      nullable=False)
    processing_operation = sa.Column(sa.String(70))
    task = sa.Column(sa.String(70), primary_key=True)
    lock_updated = sa.Column(sa.TIMESTAMP, nullable=False,
                             server_default=sa.func.now(),
                             onupdate=sa.func.now())



# Copyright (c) 2017 NEC Corp
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from neutron_lib.plugins import directory
from oslo_log import log as logging

from networking_odl.common import exceptions

LOG = logging.getLogger(__name__)

ALL_RESOURCES = {}


def get_driver(resource_type):
    try:
        return ALL_RESOURCES[resource_type]
    except KeyError:
        raise exceptions.ResourceNotRegistered(
            resource_type=resource_type) from None


class ResourceBaseDriver(object):
    """Base class for all the drivers to support full sync

       ResourceBaseDriver class acts as base class for all the drivers and
       provides default behaviour for full sync functionality.

       A driver has to provide class or object attribute RESOURCES, specifying
       resources it manages. RESOURCES must be a dictionary, keys of the
       dictionary should be resource type and value should be method suffix
       or plural used for the resources.

       A driver has to provide plugin type for itself, as class or object
       attribute. Its value should be the same, as used by neutron to
       register plugin for the resources it manages.
    """

    RESOURCES = {}
    plugin_type = None

    def __init__(self, *args, **kwargs):
        super(ResourceBaseDriver, self).__init__(*args, **kwargs)
        for resource in self.RESOURCES:
            ALL_RESOURCES[resource] = self

    def _get_resource_getter(self, method_suffix):
        method_name = "get_%s" % method_suffix
        try:
            return getattr(self.plugin, method_name)
        except AttributeError:
            raise exceptions.PluginMethodNotFound(
                plugin=self.plugin_type,
                method=method_name) from None

    def get_resources_for_full_sync(self, context, resource_type):
        """Provide all resources of type resource_type """
        if resource_type not in self.RESOURCES:
            raise exceptions.UnsupportedResourceType

        resource_getter = self._get_resource_getter(
            self.RESOURCES[resource_type])

        return resource_getter(context)

    @property
    def plugin(self):
        return directory.get_plugin(self.plugin_type)

    def get_resource_for_recovery(self, context, obj):
        resource_getter = self._get_resource_getter(obj.object_type)
        return resource_getter(context, obj.object_uuid)



#
# Copyright (C) 2016 Red Hat, Inc.
#
#  Licensed under the Apache License, Version 2.0 (the "License"); you may
#  not use this file except in compliance with the License. You may obtain
#  a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#  WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#  License for the specific language governing permissions and limitations
#  under the License.
#

from unittest import mock

import requests

from networking_l2gw.services.l2gateway.common import constants as l2gw_const
from networking_sfc.extensions import flowclassifier as fc_const
from networking_sfc.extensions import sfc as sfc_const
from neutron_lib.api.definitions import bgpvpn as bgpvpn_const
from neutron_lib.callbacks import resources
from neutron_lib.plugins import constants
from neutron_lib.plugins import directory

from networking_odl.bgpvpn import odl_v2 as bgpvpn_driver
from networking_odl.common import constants as odl_const
from networking_odl.common import exceptions
from networking_odl.db import db
from networking_odl.journal import base_driver
from networking_odl.journal import full_sync
from networking_odl.journal import journal
from networking_odl.l2gateway import driver_v2 as l2gw_driver
from networking_odl.l3 import l3_odl_v2
from networking_odl.ml2 import mech_driver_v2
from networking_odl.qos import qos_driver_v2 as qos_driver
from networking_odl.sfc.flowclassifier import sfc_flowclassifier_v2
from networking_odl.sfc import sfc_driver_v2 as sfc_driver
from networking_odl.tests import base
from networking_odl.tests.unit.journal import helper
from networking_odl.tests.unit import test_base_db
from networking_odl.trunk import trunk_driver_v2 as trunk_driver


class FullSyncTestCase(test_base_db.ODLBaseDbTestCase):
    def setUp(self):
        self.useFixture(
            base.OpenDaylightRestClientGlobalFixture(full_sync._CLIENT))
        super(FullSyncTestCase, self).setUp()

        self._CLIENT = full_sync._CLIENT.get_client()
        self.addCleanup(full_sync.FULL_SYNC_RESOURCES.clear)
        # NOTE(rajivk) workaround, Fixture defined are executed after complete
        # tests cases, but cleanup is needed after each test case.
        self.addCleanup(self._clean_registered_plugins)

    def _clean_registered_plugins(self):
        for plugin_type in self._get_all_plugins().keys():
            directory.add_plugin(plugin_type, None)

    def test_no_full_sync_when_canary_exists(self):
        full_sync.full_sync(self.db_context)
        self.assertEqual([], db.get_all_db_rows(self.db_context))

    def _filter_out_canary(self, rows):
        return [row for row in rows if row['object_uuid'] !=
                full_sync._CANARY_NETWORK_ID]

    def _mock_l2_resources(self):
        expected_journal = {odl_const.ODL_NETWORK: '1',
                            odl_const.ODL_SUBNET: '2',
                            odl_const.ODL_PORT: '3'}
        network_id = expected_journal[odl_const.ODL_NETWORK]
        plugin = mock.Mock()
        plugin.get_networks.return_value = [{'id': network_id}]
        plugin.get_subnets.return_value = [
            {'id': expected_journal[odl_const.ODL_SUBNET],
             'network_id': network_id}]
        port = {'id': expected_journal[odl_const.ODL_PORT],
                odl_const.ODL_SGS: None,
                'tenant_id': '123',
                'fixed_ips': [],
                'network_id': network_id}
        plugin.get_ports.side_effect = ([port], [])
        directory.add_plugin(constants.CORE, plugin)
        return expected_journal

    def _test_no_full_sync_when_canary_in_journal(self, state):
        self._mock_canary_missing()
        self._mock_l2_resources()
        db.create_pending_row(self.db_context, odl_const.ODL_NETWORK,
                              full_sync._CANARY_NETWORK_ID,
                              odl_const.ODL_CREATE, {})
        row = db.get_all_db_rows(self.db_context)[0]
        db.update_db_row_state(self.db_context, row, state)

        full_sync.full_sync(self.db_context)

        rows = db.get_all_db_rows(self.db_context)
        self.assertEqual([], self._filter_out_canary(rows))

    def test_no_full_sync_when_canary_pending_creation(self):
        self._test_no_full_sync_when_canary_in_journal(odl_const.PENDING)

    def test_no_full_sync_when_canary_is_processing(self):
        self._test_no_full_sync_when_canary_in_journal(odl_const.PROCESSING)

    @staticmethod
    def _get_all_resources():
        return (
            (odl_const.ODL_SG, constants.CORE),
            (odl_const.ODL_SG_RULE, constants.CORE),
            (odl_const.ODL_NETWORK, constants.CORE),
            (odl_const.ODL_SUBNET, constants.CORE),
            (odl_const.ODL_ROUTER, constants.L3),
            (odl_const.ODL_PORT, constants.CORE),
            (odl_const.ODL_FLOATINGIP, constants.L3),
            (odl_const.ODL_QOS_POLICY, constants.QOS),
            (odl_const.ODL_TRUNK, resources.TRUNK),
            (odl_const.ODL_BGPVPN, bgpvpn_const.ALIAS),
            (odl_const.ODL_BGPVPN_NETWORK_ASSOCIATION, bgpvpn_const.ALIAS),
            (odl_const.ODL_BGPVPN_ROUTER_ASSOCIATION, bgpvpn_const.ALIAS),
            (odl_const.ODL_SFC_FLOW_CLASSIFIER, fc_const.FLOW_CLASSIFIER_EXT),
            (odl_const.ODL_SFC_PORT_PAIR, sfc_const.SFC_EXT),
            (odl_const.ODL_SFC_PORT_PAIR_GROUP, sfc_const.SFC_EXT),
            (odl_const.ODL_SFC_PORT_CHAIN, sfc_const.SFC_EXT),
            (odl_const.ODL_L2GATEWAY, l2gw_const.L2GW),
            (odl_const.ODL_L2GATEWAY_CONNECTION, l2gw_const.L2GW))

    @mock.patch.object(db, 'delete_pending_rows')
    @mock.patch.object(full_sync, '_full_sync_needed')
    @mock.patch.object(full_sync, '_sync_resources')
    @mock.patch.object(journal, 'record')
    def test_sync_resource_order(
            self, record_mock, _sync_resources_mock, _full_sync_needed_mock,
            delete_pending_rows_mock):
        all_resources = self._get_all_resources()
        full_sync.FULL_SYNC_RESOURCES = {resource_type: mock.Mock()
                                         for resource_type, _ in all_resources}
        _full_sync_needed_mock._full_sync_needed.return_value = True
        context = mock.MagicMock()
        full_sync.full_sync(context)

        _sync_resources_mock.assert_has_calls(
            [mock.call(mock.ANY, object_type, mock.ANY)
                for object_type, _ in all_resources])

    def test_client_error_propagates(self):
        class TestException(Exception):
            def __init__(self):
                pass

        self._CLIENT.get.side_effect = TestException()
        self.assertRaises(TestException, full_sync.full_sync, self.db_context)

    def _mock_canary_missing(self):
        get_return = mock.MagicMock()
        get_return.status_code = requests.codes.not_found
        self._CLIENT.get.return_value = get_return

    def _assert_canary_created(self):
        rows = db.get_all_db_rows(self.db_context)
        self.assertTrue(any(r['object_uuid'] == full_sync._CANARY_NETWORK_ID
                            for r in rows))
        return rows

    def _test_full_sync_resources(self, expected_journal):
        self._mock_canary_missing()
        directory.add_plugin(constants.CORE, mock.Mock())
        full_sync.full_sync(self.db_context)

        rows = self._assert_canary_created()
        rows = self._filter_out_canary(rows)
        self.assertCountEqual(expected_journal.keys(),
                              [row['object_type'] for row in rows])
        for row in rows:
            self.assertEqual(expected_journal[row['object_type']],
                             row['object_uuid'])

    def test_full_sync_removes_pending_rows(self):
        db.create_pending_row(self.db_context, odl_const.ODL_NETWORK, "uuid",
                              odl_const.ODL_CREATE, {'foo': 'bar'})
        self._test_full_sync_resources({})

    def test_full_sync_no_resources(self):
        self._test_full_sync_resources({})

    @staticmethod
    def _get_mocked_security_groups(context):
        return [{'description': 'description',
                 'security_group_rules': ['security_grp_rules'],
                 'id': 'test_uuid', 'name': 'default'}]

    @staticmethod
    def _get_mocked_security_group_rules(context):
        return [{'direction': 'egress', 'protocol': None,
                 'description': 'description', 'port_range_max': None,
                 'id': 'test_uuid', 'security_group_id': 'test_uuid'}]

    @staticmethod
    def _get_mocked_networks(context):
        return [{'id': 'test_uuid', 'project_id': 'project_id',
                 'status': 'ACTIVE', 'subnets': [], 'description': '',
                 'name': 'network0'}]

    @staticmethod
    def _get_mocked_subnets(context):
        return [{'description': '', 'cidr': 'test-cidr', 'id': 'test_uuid',
                 'name': 'test-subnet', 'network_id': 'test_uuid',
                 'gateway_ip': 'gateway_ip'}]

    @staticmethod
    def _get_mocked_routers(context):
        return [{'status': 'ACTIVE', 'description': '', 'name': 'router1',
                 'id': 'test_uuid'}]

    @staticmethod
    def _get_mocked_ports(context):
        return [{'status': 'DOWN', 'description': None, 'id': 'test_uuid',
                 'name': 'loadbalancer-27', 'network_id': 'test_uuid',
                 'mac_address': 'fa:16:3e:69:4e:33'}]

    @staticmethod
    def _get_mocked_loadbalancers(context):
        return [{'description': '', 'tenant_id': 'tenant_id',
                 'vip_subnet_id': 'subnet_id', 'listeners': [],
                 'vip_address': '10.1.0.11', 'vip_port_id': 'port_id',
                 'pools': [], 'id': 'test_uuid', 'name': 'test-lb'}]

    @staticmethod
    def _get_mocked_listeners(context):
        return [{'admin_state_up': True, 'project_id': 'test_uuid',
                 'id': 'test_uuid'}]

    @staticmethod
    def _get_mocked_trunks(context):
        return [{'routers': [], 'id': 'test_uuid', 'name': '',
                 'tenant_id': 'project_id', 'networks': [], 'route_targets': [
                     '64512:1'], 'project_id': 'project_id', 'type': 'l3'},
                {'routers': [], 'id': 'test_uuid', 'name': '',
                 'tenant_id': 'tenant_id', 'networks': [], 'route_targets': [
                 '64512:1'], 'project_id': 'project_id', 'type': 'l3'}]

    @staticmethod
    def _get_mocked_bgpvpns(context):
        return [{'network_id': 'test_uuid', 'bgpvpn_id': 'test_uuid',
                 'project_id': 'test_uuid', 'id': 'test_uuid'}]

    @staticmethod
    def _get_mocked_l2_gateways(context):
        return [{'tenant_id': 'test_tenant_id', 'id': 'test_uuid',
                 'devices': [{'interfaces': [{'name': 'eth3'}],
                              'id': 'test_uuid', 'device_name': 'vtep0'}],
                 'name': 'test-gateway'}]

    @staticmethod
    def _get_mocked_l2_gateway_connections(context):
        return [{'network_id': 'test_uuid', 'tenant_id': 'test_uuid',
                 'l2_gateway_id': 'test_uuid', 'id': 'test_uuid'}]

    @staticmethod
    def _get_mocked_pools(context):
        return [{'name': 'pool1', 'admin_state_up': True,
                 'project_id': 'test_uuid', 'id': 'test_uuid'}]

    @staticmethod
    def _get_mocked_pool_members(context, pool_id):
        return [{'name': 'pool1', 'admin_state_up': True,
                 'project_id': 'test_uuid', 'id': 'test_uuid'}]

    @staticmethod
    def _get_mocked_healthmonitors(context):
        return [{'type': 'HTTP', 'admin_state_up': True,
                 'project_id': 'test_uuid', 'id': 'test_uuid',
                 'name': 'monitor1'}]

    @staticmethod
    def _get_mocked_listener(context):
        return [{'admin_state_up': True, 'project_id': 'test_uuid',
                 'id': 'test_uuid'}]

    @staticmethod
    def _get_mocked_floatingips(context):
        return [{'floating_network_id': 'test_uuid', 'tenant_id': 'test_uuid',
                 'dns_name': '', 'dns_domain': '', 'id': 'test_uuid'}]

    @staticmethod
    def _get_mocked_policies(context):
        return [{'id': 'test_uuid', 'project_id': 'test_uuid',
                 'name': 'test-policy', 'description': 'Policy description',
                 'shared': True, 'is_default': False}]

    @staticmethod
    def _get_mocked_bgpvpn_network_associations(context, bgpvpn_id):
        return [{'network_id': 'test_uuid', 'tenant_id': 'test_uuid',
                 'id': 'test_uuid'}]

    @staticmethod
    def _get_mocked_bgpvpn_router_associations(context, bgpvpn_id):
        return [{'router_id': 'test_uuid', 'tenant_id': 'test_uuid',
                 'id': 'test_uuid'}]

    @staticmethod
    def _get_mocked_port_chains(context):
        tenant_id = 'test_uuid'
        return [{'tenant_id': tenant_id, 'project_id': tenant_id,
                 'id': 'test_uuid'}]

    @staticmethod
    def _get_mocked_port_pair_groups(context):
        tenant_id = 'test_uuid'
        return [{'tenant_id': tenant_id, 'project_id': tenant_id,
                 'id': 'test_uuid'}]

    @staticmethod
    def _get_mocked_port_pairs(context):
        tenant_id = 'test_uuid'
        return [{'tenant_id': tenant_id, 'project_id': tenant_id,
                 'id': 'test_uuid'}]

    @staticmethod
    def _get_mocked_flowclassifiers(context):
        tenant_id = 'test_uuid'
        return [{'tenant_id': tenant_id, 'project_id': tenant_id,
                 'id': 'test_uuid'}]

    @staticmethod
    def _get_all_plugins():
        return {
            constants.CORE: (mock.Mock(),
                             mech_driver_v2.OpenDaylightMechanismDriver),
            constants.L3: (mock.Mock(), l3_odl_v2.OpenDaylightL3RouterPlugin),
            resources.TRUNK: (mock.Mock(),
                              trunk_driver.OpenDaylightTrunkHandlerV2),
            constants.QOS: (mock.Mock(), qos_driver.OpenDaylightQosDriver),
            sfc_const.SFC_EXT: (mock.Mock(),
                                sfc_driver.OpenDaylightSFCDriverV2),
            bgpvpn_const.ALIAS: (mock.Mock(),
                                 bgpvpn_driver.OpenDaylightBgpvpnDriver),
            fc_const.FLOW_CLASSIFIER_EXT: (
                mock.Mock(),
                sfc_flowclassifier_v2.OpenDaylightSFCFlowClassifierDriverV2),
            l2gw_const.L2GW: (mock.Mock(), l2gw_driver.OpenDaylightL2gwDriver)
        }

    @staticmethod
    def _get_name(resource_type):
        mapping = {
            odl_const.ODL_QOS_POLICY: odl_const.ODL_QOS_POLICIES,
            odl_const.ODL_SFC_PORT_PAIR:
                odl_const.NETWORKING_SFC_FLOW_CLASSIFIERS,
            odl_const.ODL_SFC_PORT_PAIR:
                odl_const.NETWORKING_SFC_PORT_PAIRS,
            odl_const.ODL_SFC_PORT_PAIR_GROUP:
                odl_const.NETWORKING_SFC_PORT_PAIR_GROUPS,
            odl_const.ODL_SFC_PORT_CHAIN: odl_const.NETWORKING_SFC_PORT_CHAINS,
            odl_const.ODL_L2GATEWAY_CONNECTION:
                odl_const.ODL_L2GATEWAY_CONNECTIONS}

        return ('_get_mocked_%s' % mapping.get(
            resource_type, resource_type + 's'))

    def _add_side_effect(self):
        plugins = self._get_all_plugins()
        resources = self._get_all_resources()
        for resource_type, plugin_name in resources:
            name = self._get_name(resource_type)
            setattr(plugins[plugin_name][0], "get_%s" % name[12:],
                    getattr(self, name))

            if directory.get_plugin(plugin_name) is None:
                directory.add_plugin(plugin_name, plugins[plugin_name][0])

    @mock.patch.object(journal, 'record')
    def _test_sync_resources(self, object_type, plugin_type, mocked_record):
        plugins = self._get_all_plugins()
        driver = plugins[plugin_type][1]
        args = [mock.Mock()]
        if object_type in [odl_const.ODL_BGPVPN_ROUTER_ASSOCIATION,
                           odl_const.ODL_BGPVPN_NETWORK_ASSOCIATION]:
            args.append(mock.Mock())

        resources = getattr(self, self._get_name(object_type))(*args)
        context = mock.Mock()

        def _test_get_default_handler(context, resource_type,
                                      plugin_type=plugin_type):

            resource_type = self._get_name(resource_type)[12:]
            return full_sync.get_resources(context, plugin_type=plugin_type,
                                           resource_type=resource_type)

        handler = getattr(driver, 'get_resources', _test_get_default_handler)
        full_sync._sync_resources(context, object_type, handler)
        mocked_record.assert_has_calls(
            [mock.call(context, object_type, resource['id'],
                       odl_const.ODL_CREATE,
                       resource) for resource in resources])

    def test_sync_all_resources(self):
        self._add_side_effect()
        resources = self._get_all_resources()
        for obj_type, plugin_name in resources:
            self._test_sync_resources(obj_type, plugin_name)

    def test_full_sync_retries_exceptions(self):
        with mock.patch.object(full_sync, '_full_sync_needed') as m:
            self._test_retry_exceptions(full_sync.full_sync, m)

    def test_object_not_registered(self):
        self.assertRaises(exceptions.ResourceNotRegistered,
                          full_sync.sync_resources,
                          self.db_context,
                          'test-object-type')
        self.assertEqual([], db.get_all_db_rows(self.db_context))

    def _register_resources(self):
        helper.TestDriver()
        self.addCleanup(base_driver.ALL_RESOURCES.clear)

    def add_plugin(self, plugin_type, plugin):
        directory.add_plugin(plugin_type, plugin)

    def test_plugin_not_registered(self):
        self._register_resources()
        # NOTE(rajivk): workaround, as we don't have delete method for plugin
        plugin = directory.get_plugin(helper.TEST_PLUGIN)
        directory.add_plugin(helper.TEST_PLUGIN, None)
        self.addCleanup(self.add_plugin, helper.TEST_PLUGIN, plugin)
        self.assertRaises(exceptions.PluginMethodNotFound,
                          full_sync.sync_resources,
                          self.db_context,
                          helper.TEST_RESOURCE1)
        self.assertEqual([], db.get_all_db_rows(self.db_context))

    def test_sync_resources(self):
        self._register_resources()
        plugin = helper.TestPlugin()
        self.add_plugin(helper.TEST_PLUGIN, plugin)
        resources = plugin.get_test_resource1s(self.db_context)
        full_sync.sync_resources(self.db_context,
                                 helper.TEST_RESOURCE1)
        entries = [entry.data for entry in db.get_all_db_rows(self.db_context)]
        for resource in resources:
            self.assertIn(resource, entries)
        self.assertEqual(len(resources), len(entries))

    @mock.patch.object(base_driver.ResourceBaseDriver,
                       'get_resources_for_full_sync')
    def test_get_resources_failed(self, mock_get_resources):
        self._register_resources()
        mock_get_resources.side_effect = exceptions.UnsupportedResourceType()
        resource_name = helper.TEST_RESOURCE1
        self.assertRaises(exceptions.UnsupportedResourceType,
                          full_sync.sync_resources, self.db_context,
                          resource_name)

        mock_get_resources.assert_called_once_with(self.db_context,
                                                   resource_name)

        self.assertEqual([], db.get_all_db_rows(self.db_context))



# Copyright (c) 2016 OpenStack Foundation
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import contextlib
from unittest import mock

import requests

from neutron.db import l3_db
from neutron.plugins.ml2 import plugin
from neutron.tests import base
from neutron.tests.unit.db import test_db_base_plugin_v2
from neutron.tests.unit import testlib_api
from neutron_lib.api.definitions import external_net
from neutron_lib.db import api as db_api
from neutron_lib.plugins import constants
from neutron_lib.plugins import directory
from oslo_config import cfg
from oslo_config import fixture as config_fixture
from oslo_serialization import jsonutils
from oslo_utils import uuidutils

from networking_odl.common import client
from networking_odl.common import constants as odl_const
from networking_odl.common import filters
from networking_odl.db import db
from networking_odl.journal import journal
from networking_odl.l3 import l3_odl_v2
from networking_odl.ml2 import mech_driver_v2
from networking_odl.tests import base as odl_base
from networking_odl.tests.unit import test_base_db

EMPTY_DEP = {'gw_port_id': None}
FLOATINGIP_ID = uuidutils.generate_uuid()
NETWORK_ID = uuidutils.generate_uuid()
ROUTER_ID = uuidutils.generate_uuid()
SUBNET_ID = uuidutils.generate_uuid()
PORT_ID = uuidutils.generate_uuid()


class OpenDayLightMechanismConfigTests(testlib_api.SqlTestCase):
    def setUp(self):
        self.useFixture(odl_base.OpenDaylightRestClientFixture())
        self.useFixture(odl_base.OpenDaylightFeaturesFixture())
        self.cfg = self.useFixture(config_fixture.Config())
        self.useFixture(odl_base.OpenDaylightJournalThreadFixture())
        super(OpenDayLightMechanismConfigTests, self).setUp()
        self.cfg.config(mechanism_drivers=[
                        'logger', 'opendaylight_v2'], group='ml2')
        self.cfg.config(
            port_binding_controller='legacy-port-binding', group='ml2_odl')

    def _set_config(self, url='http://127.0.0.1:9999', username='someuser',
                    password='somepass'):
        self.cfg.config(url=url, group='ml2_odl')
        self.cfg.config(username=username, group='ml2_odl')
        self.cfg.config(password=password, group='ml2_odl')

    def _test_missing_config(self, **kwargs):
        self._set_config(**kwargs)
        self.assertRaisesRegex(cfg.RequiredOptError,
                               r'value required for option \w+ in group '
                               r'\[ml2_odl\]',
                               plugin.Ml2Plugin)

    def test_valid_config(self):
        self._set_config()
        plugin.Ml2Plugin()

    def test_missing_url_raises_exception(self):
        self._test_missing_config(url=None)

    def test_missing_username_raises_exception(self):
        self._test_missing_config(username=None)

    def test_missing_password_raises_exception(self):
        self._test_missing_config(password=None)


class DataMatcher(object):

    def __init__(self, operation, object_type, object_dict):
        self._data = object_dict.copy()
        self._object_type = object_type
        filters.filter_for_odl(object_type, operation, self._data)

    def __eq__(self, s):
        data = jsonutils.loads(s)
        return self._data == data[self._object_type]

    def __ne__(self, s):
        return not self.__eq__(s)


class OpenDaylightL3TestCase(test_db_base_plugin_v2.NeutronDbPluginV2TestCase,
                             test_base_db.ODLBaseDbTestCase,
                             base.BaseTestCase):
    def setUp(self):
        self.cfg = self.useFixture(config_fixture.Config())
        self.cfg.config(core_plugin='neutron.plugins.ml2.plugin.Ml2Plugin')
        self.cfg.config(mechanism_drivers=[
                        'logger', 'opendaylight_v2'], group='ml2')
        self.useFixture(odl_base.OpenDaylightRestClientFixture())
        self.cfg.config(service_plugins=['odl-router_v2'])
        core_plugin = cfg.CONF.core_plugin
        service_plugins = {'l3_plugin_name': 'odl-router_v2'}
        self.useFixture(odl_base.OpenDaylightJournalThreadFixture())
        mock.patch.object(mech_driver_v2.OpenDaylightMechanismDriver,
                          '_record_in_journal').start()
        mock.patch.object(mech_driver_v2.OpenDaylightMechanismDriver,
                          'sync_from_callback_precommit').start()
        mock.patch.object(mech_driver_v2.OpenDaylightMechanismDriver,
                          'sync_from_callback_postcommit').start()
        self.useFixture(odl_base.OpenDaylightPeriodicTaskFixture())
        self.useFixture(odl_base.OpenDaylightFeaturesFixture())
        self.useFixture(odl_base.OpenDaylightPseudoAgentPrePopulateFixture())
        super(OpenDaylightL3TestCase, self).setUp(
            plugin=core_plugin, service_plugins=service_plugins)
        self.plugin = directory.get_plugin()
        self.plugin._network_is_external = mock.Mock(return_value=True)
        self.driver = directory.get_plugin(constants.L3)
        self.thread = journal.OpenDaylightJournalThread()

    @staticmethod
    def _get_mock_router_operation_info(network, subnet):
        router = {odl_const.ODL_ROUTER:
                  {'name': 'router1',
                   'admin_state_up': True,
                   'tenant_id': network['network']['tenant_id'],
                   'external_gateway_info': {'network_id':
                                             network['network']['id']}}}
        return router

    @staticmethod
    def _get_mock_floatingip_operation_info(network, subnet):
        floatingip = {odl_const.ODL_FLOATINGIP:
                      {'floating_network_id': network['network']['id'],
                       'tenant_id': network['network']['tenant_id'],
                       'subnet_id': None,
                       'floating_ip_address': None}}
        return floatingip

    @staticmethod
    def _get_mock_router_interface_operation_info(network, subnet):
        router_intf_dict = {'subnet_id': subnet['subnet']['id'],
                            'id': network['network']['id']}
        return router_intf_dict

    @classmethod
    def _get_mock_operation_info(cls, object_type, *args):
        getter = getattr(cls, '_get_mock_' + object_type + '_operation_info')
        return getter(*args)

    @classmethod
    def _get_mock_request_response(cls, status_code):
        response = mock.Mock(status_code=status_code)
        response.raise_for_status = mock.Mock() if status_code < 400 else (
            mock.Mock(side_effect=requests.exceptions.HTTPError(status_code)))
        return response

    def _test_operation(self, status_code, expected_calls, *args, **kwargs):
        request_response = self._get_mock_request_response(status_code)
        with mock.patch('requests.sessions.Session.request',
                        return_value=request_response) as mock_method:
            self.thread.sync_pending_entries()

        if expected_calls:
            mock_method.assert_called_with(
                headers={'Content-Type': 'application/json'},
                timeout=cfg.CONF.ml2_odl.timeout, *args, **kwargs)
        self.assertEqual(expected_calls, mock_method.call_count)

    def _call_operation_object(self, operation, object_type, object_id,
                               network, subnet):
        object_dict = self._get_mock_operation_info(
            object_type, network, subnet)
        method = getattr(self.driver, operation + '_' + object_type)

        if operation == odl_const.ODL_CREATE:
            new_object_dict = method(self.db_context, object_dict)
        elif operation == odl_const.ODL_UPDATE:
            new_object_dict = method(self.db_context, object_id, object_dict)
        else:
            new_object_dict = method(self.db_context, object_id)

        return new_object_dict

    def _test_operation_thread_processing(self, object_type, operation,
                                          network, subnet, object_id,
                                          expected_calls=1):
        http_requests = {odl_const.ODL_CREATE: 'post',
                         odl_const.ODL_UPDATE: 'put',
                         odl_const.ODL_DELETE: 'delete'}
        status_codes = {odl_const.ODL_CREATE: requests.codes.created,
                        odl_const.ODL_UPDATE: requests.codes.ok,
                        odl_const.ODL_DELETE: requests.codes.no_content}

        http_request = http_requests[operation]
        status_code = status_codes[operation]

        # Create database entry.
        new_object_dict = self._call_operation_object(
            operation, object_type, object_id, network, subnet)

        # Setup expected results.
        if operation in [odl_const.ODL_UPDATE, odl_const.ODL_DELETE]:
            url = (cfg.CONF.ml2_odl.url + '/' + object_type + 's/' +
                   object_id)
        else:
            url = cfg.CONF.ml2_odl.url + '/' + object_type + 's'

        if operation in [odl_const.ODL_CREATE, odl_const.ODL_UPDATE]:
            kwargs = {
                'url': url,
                'data': DataMatcher(operation, object_type, new_object_dict)}
        else:
            kwargs = {'url': url, 'data': None}

        # Call threading routine to process database entry. Test results.
        self._test_operation(status_code, expected_calls, http_request,
                             **kwargs)

        return new_object_dict

    def _test_thread_processing(self, object_type):
        # Create network and subnet.
        kwargs = {'arg_list': (external_net.EXTERNAL,),
                  external_net.EXTERNAL: True}
        with self.network(**kwargs) as network:
            with self.subnet(network=network, cidr='10.0.0.0/24'):
                # Add and process create request.
                new_object_dict = self._test_operation_thread_processing(
                    object_type, odl_const.ODL_CREATE, network, None, None)
                object_id = new_object_dict['id']
                rows = db.get_all_db_rows_by_state(self.db_context,
                                                   odl_const.COMPLETED)
                self.assertEqual(1, len(rows))

                # Add and process 'update' request. Adds to database.
                self._test_operation_thread_processing(
                    object_type, odl_const.ODL_UPDATE, network, None,
                    object_id)
                rows = db.get_all_db_rows_by_state(self.db_context,
                                                   odl_const.COMPLETED)
                self.assertEqual(2, len(rows))

                # Add and process 'delete' request. Adds to database.
                self._test_operation_thread_processing(
                    object_type, odl_const.ODL_DELETE, network, None,
                    object_id)
                rows = db.get_all_db_rows_by_state(self.db_context,
                                                   odl_const.COMPLETED)
                self.assertEqual(3, len(rows))

    def _test_db_results(self, object_id, operation, object_type):
        rows = db.get_all_db_rows(self.db_context)

        self.assertEqual(1, len(rows))
        self.assertEqual(operation, rows[0]['operation'])
        self.assertEqual(object_type, rows[0]['object_type'])
        self.assertEqual(object_id, rows[0]['object_uuid'])

        self._db_cleanup()

    @contextlib.contextmanager
    def _prepare_resource(self, resource_type):
        # Create network and subnet for testing.
        kwargs = {'arg_list': (external_net.EXTERNAL,),
                  external_net.EXTERNAL: True}
        with self.network(**kwargs) as network:
            with self.subnet(network=network):
                yield self._get_mock_operation_info(
                    resource_type, network, None)

    def _test_object_db(self, object_type):
        with self._prepare_resource(object_type) as object_dict:
            # Add and test 'create' database entry.
            method = getattr(self.driver,
                             odl_const.ODL_CREATE + '_' + object_type)
            new_object_dict = method(self.db_context, object_dict)
            object_id = new_object_dict['id']
            self._test_db_results(object_id, odl_const.ODL_CREATE, object_type)

            # Add and test 'update' database entry.
            method = getattr(self.driver,
                             odl_const.ODL_UPDATE + '_' + object_type)
            with db_api.CONTEXT_WRITER.using(self.db_context):
                method(self.db_context, object_id, object_dict)
                self._test_db_results(
                    object_id, odl_const.ODL_UPDATE, object_type)

            # Add and test 'delete' database entry.
            method = getattr(self.driver,
                             odl_const.ODL_DELETE + '_' + object_type)
            method(self.db_context, object_id)
            self._test_db_results(object_id, odl_const.ODL_DELETE, object_type)

    def _test_dependency_processing(
            self, test_operation, test_object, test_id, test_data,
            dep_operation, dep_object, dep_id, dep_data):

        # Mock sendjson to verify that it never gets called.
        mock_sendjson = mock.patch.object(client.OpenDaylightRestClient,
                                          'sendjson').start()

        # Create dependency db row and mark as 'processing' so it won't
        # be processed by the journal thread.
        ctxt = self.db_context
        with db_api.CONTEXT_WRITER.using(self.db_context):
            journal.record(ctxt, dep_object, dep_id, dep_operation, dep_data)
            row = db.get_all_db_rows_by_state(self.db_context,
                                              odl_const.PENDING)
            db.update_db_row_state(self.db_context, row[0],
                                   odl_const.PROCESSING)

            # Create test row with dependent ID.
            journal.record(ctxt, test_object, test_id, test_operation,
                           test_data)

        # Call journal thread.
        self.thread.sync_pending_entries()

        # Verify that dependency row is still set at 'processing'.
        rows = db.get_all_db_rows_by_state(self.db_context,
                                           odl_const.PROCESSING)
        self.assertEqual(1, len(rows))

        # Verify that the test row was processed and set back to 'pending'
        # to be processed again.
        rows = db.get_all_db_rows_by_state(self.db_context, odl_const.PENDING)
        self.assertEqual(1, len(rows))

        # Verify that _json_data was not called.
        self.assertFalse(mock_sendjson.call_count)

    def test_router_db(self):
        self._test_object_db(odl_const.ODL_ROUTER)

    def test_floatingip_db(self):
        self._test_object_db(odl_const.ODL_FLOATINGIP)

    def test_router_threading(self):
        self._test_thread_processing(odl_const.ODL_ROUTER)

    def test_floatingip_threading(self):
        self._test_thread_processing(odl_const.ODL_FLOATINGIP)

    def test_delete_network_validate_ext_delete_router_dep(self):
        router_context = [NETWORK_ID]
        self._test_dependency_processing(
            odl_const.ODL_DELETE, odl_const.ODL_NETWORK, NETWORK_ID, None,
            odl_const.ODL_DELETE, odl_const.ODL_ROUTER, ROUTER_ID,
            router_context)

    def test_create_router_validate_ext_create_port_dep(self):
        router_context = {'gw_port_id': PORT_ID}
        self._test_dependency_processing(
            odl_const.ODL_CREATE, odl_const.ODL_ROUTER, ROUTER_ID,
            router_context,
            odl_const.ODL_CREATE, odl_const.ODL_PORT, PORT_ID,
            {'fixed_ips': [], 'network_id': None, odl_const.ODL_SGS: None,
             'tenant_id': 'tenant'})

    def test_delete_router_validate_ext_delete_floatingip_dep(self):
        floatingip_context = [ROUTER_ID]
        self._test_dependency_processing(
            odl_const.ODL_DELETE, odl_const.ODL_ROUTER, ROUTER_ID, None,
            odl_const.ODL_DELETE, odl_const.ODL_FLOATINGIP, FLOATINGIP_ID,
            floatingip_context)

    def test_delete_router_validate_self_create_dep(self):
        self._test_dependency_processing(
            odl_const.ODL_DELETE, odl_const.ODL_ROUTER, ROUTER_ID, EMPTY_DEP,
            odl_const.ODL_CREATE, odl_const.ODL_ROUTER, ROUTER_ID, EMPTY_DEP)

    def test_delete_router_validate_self_update_dep(self):
        self._test_dependency_processing(
            odl_const.ODL_DELETE, odl_const.ODL_ROUTER, ROUTER_ID, EMPTY_DEP,
            odl_const.ODL_UPDATE, odl_const.ODL_ROUTER, ROUTER_ID, EMPTY_DEP)

    def test_update_router_validate_self_create_dep(self):
        self._test_dependency_processing(
            odl_const.ODL_UPDATE, odl_const.ODL_ROUTER, ROUTER_ID, EMPTY_DEP,
            odl_const.ODL_CREATE, odl_const.ODL_ROUTER, ROUTER_ID, EMPTY_DEP)

    def test_create_floatingip_validate_ext_create_network_dep(self):
        floatingip_context = {'floating_network_id': NETWORK_ID}
        self._test_dependency_processing(
            odl_const.ODL_CREATE, odl_const.ODL_FLOATINGIP, FLOATINGIP_ID,
            floatingip_context,
            odl_const.ODL_CREATE, odl_const.ODL_NETWORK, NETWORK_ID, {})

    def test_update_floatingip_validate_self_create_dep(self):
        floatingip_context = {'floating_network_id': NETWORK_ID}
        self._test_dependency_processing(
            odl_const.ODL_UPDATE, odl_const.ODL_FLOATINGIP, FLOATINGIP_ID,
            floatingip_context,
            odl_const.ODL_CREATE, odl_const.ODL_FLOATINGIP, FLOATINGIP_ID,
            EMPTY_DEP)

    def test_delete_floatingip_validate_self_create_dep(self):
        self._test_dependency_processing(
            odl_const.ODL_DELETE, odl_const.ODL_FLOATINGIP, FLOATINGIP_ID,
            EMPTY_DEP,
            odl_const.ODL_CREATE, odl_const.ODL_FLOATINGIP, FLOATINGIP_ID,
            {})

    def test_delete_floatingip_validate_self_update_dep(self):
        self._test_dependency_processing(
            odl_const.ODL_DELETE, odl_const.ODL_FLOATINGIP, FLOATINGIP_ID,
            EMPTY_DEP,
            odl_const.ODL_UPDATE, odl_const.ODL_FLOATINGIP, FLOATINGIP_ID,
            {})

    @mock.patch.object(journal, 'record')
    def test__record_in_journal_retries(self, record_mock):
        self._test_retry_exceptions(
            l3_odl_v2._record_in_journal, record_mock, True)

    def _assert_record_in_journal(self, record_in_journal, resource_type,
                                  operation):
        record_in_journal.assert_called_with(
            mock.ANY, resource_type, operation, mock.ANY, mock.ANY)

    def _call_and_assert_recorded_in_journal(
            self, resource_type, operation, function, *args):
        with mock.patch.object(l3_odl_v2,
                               '_record_in_journal') as record_in_journal:
            function(self.db_context, *args)

        record_in_journal.assert_called_with(
            mock.ANY, resource_type, operation, mock.ANY, mock.ANY)

    def test_create_router_records_in_journal(self):
        with self._prepare_resource(odl_const.ODL_ROUTER) as router:
            self._call_and_assert_recorded_in_journal(
                odl_const.ODL_ROUTER, odl_const.ODL_CREATE,
                self.driver.create_router, router)

    def test_update_router_records_in_journal(self):
        with self._prepare_resource(odl_const.ODL_ROUTER) as router:
            result = self.driver.create_router(self.db_context, router)
            self._call_and_assert_recorded_in_journal(
                odl_const.ODL_ROUTER, odl_const.ODL_UPDATE,
                self.driver.update_router, result['id'], router)

    def test_delete_router_records_in_journal(self):
        with self._prepare_resource(odl_const.ODL_ROUTER) as router:
            result = self.driver.create_router(self.db_context, router)
            self._call_and_assert_recorded_in_journal(
                odl_const.ODL_ROUTER, odl_const.ODL_DELETE,
                self.driver.delete_router, result['id'])

    def test_create_fip_records_in_journal(self):
        with self._prepare_resource(odl_const.ODL_FLOATINGIP) as fip:
            self._call_and_assert_recorded_in_journal(
                odl_const.ODL_FLOATINGIP, odl_const.ODL_CREATE,
                self.driver.create_floatingip, fip)

    def test_update_fip_records_in_journal(self):
        with self._prepare_resource(odl_const.ODL_FLOATINGIP) as fip:
            result = self.driver.create_floatingip(self.db_context, fip)
            self._call_and_assert_recorded_in_journal(
                odl_const.ODL_FLOATINGIP, odl_const.ODL_UPDATE,
                self.driver.update_floatingip, result['id'], fip)

    def test_delete_fip_records_in_journal(self):
        with self._prepare_resource(odl_const.ODL_FLOATINGIP) as fip:
            result = self.driver.create_floatingip(self.db_context, fip)
            self._call_and_assert_recorded_in_journal(
                odl_const.ODL_FLOATINGIP, odl_const.ODL_DELETE,
                self.driver.delete_floatingip, result['id'])

    @mock.patch.object(l3_db.L3_NAT_dbonly_mixin, 'disassociate_floatingips')
    @mock.patch.object(l3_odl_v2.OpenDaylightL3RouterPlugin, 'get_floatingips')
    def test_disassociate_floatingips_records_in_journal(
            self, get_fips, disassociate_floatingips):
        with self._prepare_resource(odl_const.ODL_FLOATINGIP) as fip:
            result = self.driver.create_floatingip(self.db_context, fip)
            get_fips.return_value = [result]
            self._call_and_assert_recorded_in_journal(
                odl_const.ODL_FLOATINGIP, odl_const.ODL_UPDATE,
                self.driver.disassociate_floatingips, 'fake_id')

        self.assertTrue(disassociate_floatingips.called)



# Copyright (C) 2017 Red Hat Inc.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
#

"""Added version_id for optimistic locking

Revision ID: 43af357fd638
Revises: 3d560427d776
Create Date: 2016-03-24 10:14:56.408413

"""

from alembic import op
import sqlalchemy as sa

# revision identifiers, used by Alembic.
revision = '43af357fd638'
down_revision = '3d560427d776'
depends_on = ('fa0c536252a5',)


def upgrade():
    op.add_column('opendaylightjournal',
                  sa.Column('version_id', sa.Integer, server_default='0',
                            nullable=False))
